;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// TUTPROG2.CPP - VGA Trainer Program 2 (in Turbo C++ 3.0)                 //
;//                                                                         //
;// "The VGA Trainer Program" is written by Denthor of Asphyxia. However it //
;// was limited to Pascal only in its first run.  All I have done is taken  //
;// his original release, translated it to C++ and touched up a few things. //
;// I take absolutely no credit for the concepts presented in this code and //
;// am NOT the person to ask for help if you are having trouble.            //
;//                                                                         //
;// Program Notes : This program presents many new concepts, including:     //
;//                 line drawing, pallette manipulation, and fading.        //
;//                                                                         //
;//                 If you are compiling this code command line, be sure to //
;//                 use the "-ml" parameter (large memory model).           //
;//                 Otherwise, the program will compile and link, but will  //
;//                 lock up your system.                                    //
;//                                                                         //
;// Author        : Grant Smith (Denthor) - denthor@beastie.cs.und.ac.za    //
;// Translator    : Christopher G. Mann   - r3cgm@dax.cc.uakron.edu         //
;//                                                                         //
;// Last Modified : November 29, 1994                                       //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

;*****************************************************************************/
;*                                                                           */
;*         Translated to ASM by Robert Cassels (Ossa) : June 2006            */
;*                                                                           */
;*****************************************************************************/

.MODEL SMALL, C
.386
.DOSSEG
.STACK 200h

;//               //
;// INCLUDE FILES //
;//               //

INCLUDE CEQUIV.INC

;//                     //
;// FUNCTION PROTOTYPES //
;//                     //

;// MODE SETTING FUNCTIONS

SetMCGA				PROTO
SetText				PROTO

;// PALLETTE FUNCTIONS
GrabPallette		PROTO
GetPal				PROTO ColorNo:BYTE, pR:WORD, pG:WORD, pB:WORD
Pal					PROTO ColorNo:BYTE,  R:BYTE,  G:BYTE,  B:BYTE
PalPlay				PROTO
Fadeup				PROTO
FadeDown			PROTO
Blackout			PROTO
RestorePallette		PROTO

;// SMALL UTILITY FUNCTIONS
sgn					PROTO a:WORD ; (DOWRD? REAL4?)
round				PROTO a:WORD ; (DOWRD? REAL4?)

;// DRAWING FUNCTIONS
Putpixel			PROTO x:WORD, y:WORD, Col:BYTE
Line				PROTO a:WORD, b:WORD, _c:WORD, d:WORD, col:BYTE
WaitRetrace			PROTO

;// MID-LEVEL (COMBINATION) FUNCTIONS
SetUpScreen			PROTO
HiddenScreenSetup	PROTO

;//                              //
;// GLOBAL VARIABLE DECLARATIONS //
;//                              //

.DATA?

;// This declares the PALL variable. 0 to 255 signifies the colors of the
;// pallette, 1 to 3 signifies the Red, Green and Blue values. I am
;// going to use this as a sort of "virtual pallette", and alter it
;// as much as I want, then suddenly bang it to screen. Pall2 is used
;// to "remember" the origional pallette so that we can restore it at
;// the end of the program. */

pall_element STRUCT
	R				BYTE				?
	G				BYTE				?
	B				BYTE				?
pall_element ENDS

Pall				pall_element		256 dup (<>)
Pall2				pall_element		256 dup (<>)
Tmp					pall_element		<>

.DATA

szBegin				db					"This program will draw lines of different colors across the", 0Dh, 0Ah
					db					"screen and change them only by changing their pallette values.", 0Dh, 0Ah
					db					"The nice thing about using the pallette is that one pallette", 0Dh, 0Ah
					db					"change changes the same color over the whole screen, without", 0Dh, 0Ah
					db					"you having to redraw it. Because I am using a WaitRetrace", 0Dh, 0Ah
					db					"command, turning on and off your turbo during the demonstration", 0Dh, 0Ah
					db					"should have no effect.", 0Dh, 0Ah, 0Dh, 0Ah
					db					"The second part of the demo blacks out the screen using the", 0Dh, 0Ah
					db					"pallette, fades in the screen, waits for a keypress, then fades", 0Dh, 0Ah
					db					"it out again. I haven''t put in any delays for the fadein/out,", 0Dh, 0Ah
					db					"so you will have to put them in yourself to get to the speed", 0Dh, 0Ah
					db					"you like. Have fun and enjoy! ;-)", 0Dh, 0Ah, 0Dh, 0Ah, 0Dh, 0Ah
					db					"Hit any key to continue ...", 0

szEnd				db					"All done. This concludes the second sample program in the ASPHYXIA", 0Dh, 0Ah
					db					"Training series. You may reach DENTHOR under the name of GRANT", 0Dh, 0Ah
					db					"SMITH on the MailBox BBS, or leave a message to ASPHYXIA on the", 0Dh, 0Ah
					db					"ASPHYXIA BBS. Get the numbers from Roblist, or write to :", 0Dh, 0Ah
					db					"             Grant Smith", 0Dh, 0Ah
					db					"             P.O. Box 270", 0Dh, 0Ah
					db					"             Kloof", 0Dh, 0Ah
					db					"             3640", 0Dh, 0Ah
					db					"I hope to hear from you soon!", 0Dh, 0Ah, 0Dh, 0Ah, 0Dh, 0Ah
					db					"Hit any key to exit ...", 0

;///////////////////////////////////////////////////////////////////////////////
;//                                                                           //
;//                                 ENTRY POINT                               //
;//                                                                           //
;///////////////////////////////////////////////////////////////////////////////

.CODE

	.STARTUP
	
	invoke clrscr
	invoke outstr, addr szBegin
	invoke getch

	invoke SetMCGA
	invoke GrabPallette
	invoke SetUpScreen
	
@@:
	invoke PalPlay

	invoke kbhit
	test ax, ax
	jz @B

	invoke getch	; once a key is pressed, be sure to clear the keyboard buffer
	
	invoke Blackout
	invoke HiddenScreenSetup
	invoke Fadeup
	invoke getch
	invoke FadeDown
	invoke getch
	
	invoke RestorePallette
	invoke SetText
	invoke outstr, addr szEnd
	invoke getch
	
	.EXIT

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// SetMCGA() - This function gets you into 320x200x256 mode.               //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

SetMCGA PROC
	mov ax, 0013h
	int 10h
	ret
SetMCGA ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// SetText() - This function gets you into text mode.                      //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

SetText PROC
	mov ax, 0003h
	int 10h
	ret
SetText ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// GrabPallette() - This function retrieves the current pallette.          //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

GrabPallette PROC
	LOCAL bCur:BYTE

	mov ax, DGROUP
	mov es, ax
	mov ds, ax

	mov bCur, 0
GrabPallette_loop:

	xor ax, ax
	mov al, bCur
	mov bx, ax
	shl bx, 1
	add bx, ax
	add bx, offset Pall2

	invoke GetPal, bCur, addr [bx][pall_element.R], addr [bx][pall_element.G], addr [bx][pall_element.B]

	add bCur, 1
	jnz GrabPallette_loop

	ret
GrabPallette ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// GetPal() - This reads the values of the Red, Green, and Blue values of  //
;//            a certain color.  This function uses pass-by-reference.      //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

GetPal PROC ColorNo:BYTE, pR:WORD, pG:WORD, pB:WORD

	mov ax, DGROUP
	mov ds, ax

	mov dx, 03C7h
	mov al, ColorNo
	out dx, al

	mov dx, 03C9h

	mov bx, pR
	in al, dx
	mov ds:[bx], al

	mov bx, pG
	in al, dx
	mov ds:[bx], al

	mov bx, pB
	in al, dx
	mov ds:[bx], al

	ret
GetPal ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// Pal() - This sets the Red, Green, and Blue values of a certain color.   //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

Pal PROC ColorNo:BYTE, R:BYTE, G:BYTE, B:BYTE

	mov dx, 03C8h
	mov al, ColorNo
	out dx, al

	mov dx, 03C9h
	mov al, R
	out dx, al

	mov dx, 03C9h
	mov al, G
	out dx, al

	mov dx, 03C9h
	mov al, B
	out dx, al

	ret
Pal ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// PalPlay() - This function mucks about with our "virtual pallette", then //
;//             shoves it to the screen.                                    //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////


PalPlay PROC
	LOCAL bCol:BYTE
	
	mov ax, DGROUP
	mov es, ax
	mov ds, ax
	
	; es:[di]
	; ds:[si]
	
	; memmove(Tmp,Pall[1],3);
	
	mov cx, 3
	mov si, (offset Pall) + (1 * sizeof(pall_element))
	mov di, offset Tmp
	rep movsb

	; memmove(Pall[1],Pall[2],199*3);
	
	mov cx, 199 * sizeof(pall_element)
	mov si, (offset Pall) + (2 * sizeof(pall_element))
	mov di, (offset Pall) + (1 * sizeof(pall_element))
	rep movsb

	; memmove(Pall[200],Tmp,3);
	
	mov cx, 3
	mov di, (offset Pall) + (200 * sizeof(pall_element))
	mov si, offset Tmp
	rep movsb

	invoke WaitRetrace
	
	mov bCol, 0
	mov bx, offset Pall
@@:
	push bx
	invoke Pal, bCol, [bx][pall_element.R], [bx][pall_element.G], [bx][pall_element.B]
	pop bx

	add bx, sizeof pall_element
	add bCol, 1
	jnz @B
	
	ret
PalPlay ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// Fadeup() - This function slowly fades up the new screen                 //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

Fadeup PROC
	LOCAL loop1:BYTE
	LOCAL loop2:BYTE
	LOCAL pe:pall_element
	
	; A color value for Red, green or blue is 0 to 63, so this loop only
	; need be executed a maximum of 64 times.

	mov loop1, 0
Fadeup_loop1:

	invoke WaitRetrace

	mov loop2, 0
Fadeup_loop2:

	invoke GetPal, loop2, offset Tmp.R, offset Tmp.G, offset Tmp.B

	; bx = offset Pall2[loop2][0]

	mov ax, DGROUP
	mov es, ax
	mov ds, ax

	xor ax, ax
	mov al, loop2
	mov bx, ax
	shl bx, 1
	add bx, ax
	add bx, offset Pall2
	
	mov al, ds:[bx+0]
	cmp al, Tmp.R
	jbe @F
	cmp Tmp.R, 63
	jnb @F
	add Tmp.R, 1
@@:
	
	mov al, ds:[bx+1]
	cmp al, Tmp.G
	jbe @F
	cmp Tmp.G, 63
	jnb @F
	add Tmp.G, 1
@@:
	
	mov al, ds:[bx+2]
	cmp al, Tmp.B
	jbe @F
	cmp Tmp.B, 63
	jnb @F
	add Tmp.B, 1
@@:
	
	invoke Pal, loop2, Tmp.R, Tmp.G, Tmp.B

	add loop2, 1
	jnz Fadeup_loop2

	add loop1, 1
	cmp loop1, 64
	jne Fadeup_loop1
	
	ret
Fadeup ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// FadeDown() - This function fades the screen out to black.               //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////
;
FadeDown PROC
	LOCAL loop1:BYTE
	LOCAL loop2:BYTE
	LOCAL pe:pall_element
	
	; A color value for Red, green or blue is 0 to 63, so this loop only
	; need be executed a maximum of 64 times.

	mov loop1, 0
FadeDown_loop1:

	invoke WaitRetrace

	mov loop2, 0
FadeDown_loop2:

	invoke GetPal, loop2, offset Tmp.R, offset Tmp.G, offset Tmp.B

	or Tmp.R, 0
	jz @F
	sub Tmp.R, 1
@@:
	
	or Tmp.G, 0
	jz @F
	sub Tmp.G, 1
@@:
	
	or Tmp.B, 0
	jz @F
	sub Tmp.B, 1
@@:
	
	invoke Pal, loop2, Tmp.R, Tmp.G, Tmp.B

	add loop2, 1
	jnz FadeDown_loop2

	add loop1, 1
	cmp loop1, 64
	jne FadeDown_loop1
	
	ret
FadeDown ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// Blackout() - This function just clears the screen.                      //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

Blackout PROC
	LOCAL bCol:BYTE

	invoke WaitRetrace
	
	mov bCol, 0
@@:
	invoke Pal, bCol, 0, 0, 0
	
	add bCol, 1
	jnz @B
	
	ret
Blackout ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// RestorePallette() - This function restores the origional pallette.      //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

RestorePallette PROC
	LOCAL bCur:BYTE

	mov ax, DGROUP
	mov es, ax
	mov ds, ax

	mov bCur, 0
RestorePallette_loop:

	xor ax, ax
	mov al, bCur
	mov bx, ax
	shl bx, 1
	add bx, ax
	add bx, offset Pall2

	invoke Pal, bCur, [bx][pall_element.R], [bx][pall_element.G], [bx][pall_element.B]

	add bCur, 1
	jnz RestorePallette_loop

	ret
RestorePallette ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// sgn() - This function is used by Line() to determine the sign of a long //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

sgn PROC a:WORD

	mov ax, a
	cmp ax, 0
	jl  sgn_neg
	je  sgn_zero
	mov ax, 1
	ret
sgn_neg:
	mov ax, 0FFFFh
sgn_zero:
	ret
sgn ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// Putpixel() - This puts a pixel on the screen by writing directly to     //
;//              memory.                                                    //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

Putpixel PROC x:WORD, y:WORD, Col:BYTE
	mov ax, 0A000h
	mov es, ax
	
	mov ax, y
	mov cx, 320
	mul cx
	mov bx, ax
	add bx, x
	mov dl, Col
	mov es:[bx], dl
	
	ret
Putpixel ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// Line() - This draws a line from a,b to c,d of color col.                //
;//          This function will be explained in more detail in tut3new.zip  //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

Line PROC a:WORD, b:WORD, _c:WORD, d:WORD, col:BYTE
	LOCAL u:WORD, s:WORD, v:WORD, m:WORD, n:WORD
	LOCAL d1x:WORD, d1y:WORD, d2x:WORD, d2y:WORD
	LOCAL i:WORD

	; u = x2 - x1

	mov ax, _c
	sub ax, a
	mov u, ax

	; v = y2 - y1

	mov ax, d
	sub ax, b
	mov v, ax

	; d1x is the sign of u (x2-x1) (VALUE -1,0,1)

	invoke sgn, u
	mov d1x, ax

	; d1y is the sign of v (y2-y1) (VALUE -1,0,1)

	invoke sgn, v
	mov d1y, ax

	; d2x is the sign of u (x2-x1) (VALUE -1,0,1)

	invoke sgn, u
	mov d2x, ax
	mov d2y, 0

	; n is the distance between y1 and y2
	
	invoke abs, v
	mov n, ax

	; m is the distance between x1 and x2

	invoke abs, u
	mov m, ax
	
	; if the x distance is greater than the y distance

	.if ax <= n
		; swap n <-> m
		mov dx, n
		mov m, dx
		mov n, ax
		
		; swap 
		
		mov d2x, 0
		
		; d2y is the sign of v (x2-x1) (VALUE -1,0,1)
		
		invoke sgn, v
		mov d2y, ax
	.endif

	; s is the m distance (either x or y) divided by 2

	mov ax, m
	shr ax, 1
	mov s, ax

	; repeat this loop until it
	; is = to m (y or x distance)

	mov i, 0
Line_loop:
	
	invoke PutPixel, a, b, col
	
	mov ax, s
	add ax, n
	
	.if ax >= m
		sub ax, m
		mov s, ax
		
		mov ax, a
		add ax, d1x
		mov a, ax
		
		mov ax, b
		add ax, d1y
		mov b, ax
	.else
		mov s, ax

		mov ax, a
		add ax, d2x
		mov a, ax
		
		mov ax, b
		add ax, d2y
		mov b, ax
	.endif
	
	mov ax, i
	add ax, 1
	mov i, ax
	cmp ax, m
	jbe Line_loop
	
	ret
Line ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// WaitRetrace() - This waits until you are in a Verticle Retrace ...      //
;//                 this means that all screen manipulation you do only     //
;//                 appears on screen in the next verticle retrace ... this //
;//                 removes most of the "fuzz" that you see on the screen   //
;//                 when changing the pallette. It unfortunately slows down //
;//                 your program by "synching" your program with your       //
;//                 monitor card ... it does mean that the program will run //
;//                 at almost the same speed on different speeds of         //
;//                 computers which have similar monitors.  In our          //
;//                 SilkyDemo, we used a WaitRetrace, and it therefore      //
;//                 runs at the same (fairly fast) speed when Turbo is on   //
;//                 or off.                                                 //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

WaitRetrace PROC
	mov dx, 03DAh

@@:
	in  al, dx
	and al, 08h
	jnz @B

@@:
	in  al, dx
	and al, 08h
	jz  @B

	ret
WaitRetrace ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// SetUpScreen() - This gets our screen ready by setting up the pallette   //
;//                 and drawing the lines.                                  //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

SetUpScreen PROC
	LOCAL wCol:WORD

	mov ax, DGROUP
	mov es, ax
	mov ds, ax
	
	; Clear the entire PALL variable to zero.
	; memset(Pall,0,sizeof(Pall));
	
	mov cx, sizeof Pall
	xor al, al
	mov di, offset Pall
	rep stosb
	
	; This sets colors 0 to 200 in the PALL variable to values between
	; 0 to 63. the modulus function gives you the remainder of a division,
	; ie. 105 % 10 = 5.
	
	; for (int loop = 0; loop < 201;loop++)
	;  Pall[loop][0] = loop % 64;
	
	mov cx, 0
	mov bx, offset Pall
@@:
	mov ax, cx
	and ax, 003Fh
	
	mov ds:[bx], al
	
	add bx, sizeof pall_element
	add cx, 1
	cmp cx, 201
	jne @B
	
	mov wCol, 0
SetUpScreen_loop:
	
	; Line(319,199,320-loop,0,(loop % 199) + 1)
	
	mov ax, wCol
	xor dx, dx
	mov cx, 319
	sub cx, ax
	mov bx, 199
	div bx
	mov ax, 199
	sub ax, dx
	add ax, 1
	
	push ax
	
	invoke Line, 319, 199, cx, 0, al
	
	; Line(0,0,loop,199,(loop % 199) + 1);
	
	pop ax
	
	invoke Line, 0, 0, wCol, 199, al
	
	invoke PalPlay
	
	add wCol, 1
	cmp wCol, 320
	jne SetUpScreen_loop
	
	ret
SetUpScreen ENDP

;/////////////////////////////////////////////////////////////////////////////
;//                                                                         //
;// HiddenScreenSetup() - This function sets up the screen while it is      //
;//                       blacked out, so that the user can't see what is   //
;//                       happening.                                        //
;//                                                                         //
;/////////////////////////////////////////////////////////////////////////////

HiddenScreenSetup PROC
	LOCAL loop1:WORD
	LOCAL loop2:WORD

	mov loop1, 0
TestINTPutpixel_loop1:

	mov loop2, 0
TestINTPutpixel_loop2:
	
	invoke rand
	invoke Putpixel, loop1, loop2, al
	
	add loop2, 1
	cmp loop2, 200
	jne TestINTPutpixel_loop2

	add loop1, 1
	cmp loop1, 320
	jne TestINTPutpixel_loop1
	
	ret
HiddenScreenSetup ENDP

END